/*
** LocalRodeo - Client Side Protection against JavaScript Malware
** Copyright (C) 2006 Justus Winter <4winter@informatik.uni-hamburg.de>
**                and Martin Johns <johns@informatik.uni-hamburg.de>
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 2
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
** MA  02110-1301, USA
*/

/***********************************************************
constants
***********************************************************/

// reference to the interface defined in LocalRodeo-nsIContentPolicy.xpt
const nsILocalRodeo = Components.interfaces.nsIContentPolicy;

// reference to the required base interface that all components must support
const nsISupports = Components.interfaces.nsISupports;
const nsIObserver = Components.interfaces.nsIObserver;
const nsIContentPolicy = Components.interfaces.nsIContentPolicy;

// UUID uniquely identifying our component
const CLASS_ID = Components.ID('{31f37a36-5dd4-4f92-babd-30457aa09bb0}');

// description
const CLASS_NAME = 'LocalRodeo XPCOM Component implementing nsIContentPolicy';

// textual unique identifier
const CONTRACT_ID = '@requestrodeo/contentpolicy;1';

/****
Globals
*****/

var ro;
var antiPinnerList = new Array();
var headerList = [ "Status", "Date", "Server", "Last-Modified", "Etag", "Accept-Ranges", "Content-Length", "Keep-Alive", "Connection", "Content-Type", "Content-Location", "Vary", "TCN", "Content-Language"];

/***********************************************************
class definition
***********************************************************/

function RodeoObserver(rr){
    this["rr"] = rr;
    antiPinnerList = new Array();
}

RodeoObserver.prototype = {
    /* from TamperData */
  _observerService: Components.classes["@mozilla.org/observer-service;1"].getService(Components.interfaces.nsIObserverService),
  addToListener: function()
  {
    //this._observerService.addObserver(this, "http-on-modify-request",   false);
    this._observerService.addObserver(this, "http-on-examine-response", false);
    logToConsole("LocalRodeo enabled.");
  },

  /* from TamperData */
  removeFromListener: function()
  {
    //this._observerService.removeObserver(this, "http-on-modify-request");
    this._observerService.removeObserver(this, "http-on-examine-response");
    logToConsole("LocalRodeo disabled.");
  },
  
  /* This is the cut down observerService's observe listener from TamperData */
  observe: function(aSubject, aTopic, aData){
    if (!this.rr._enabled()){
        return 0;
    }
  
    if (aTopic == 'http-on-examine-response'){
	  aSubject.QueryInterface(Components.interfaces.nsIHttpChannel);
	  this.onExamineResponse(aSubject);
    } 
  },
  
  // Check for redirect (30x) resposes, as they are nor catched by nsIContentPolicy
  onExamineResponse: function(httpChannel){
  
    // check for requests to hosts that are marked to be part of a anti-pinning attack
    host = this.rr.getHostname(httpChannel.URI.asciiSpec);
    if (antiPinnerList[host]){
	//debugMsg("Stop anti-pinner\n");
       for (i in headerList){
            try{
                httpChannel.setResponseHeader(headerList[i],"",false);
            } catch (e){
                //debugMsg("Header kill failed: " + headerList[i] + "\n");
                //debugMsg(e + "\n")
            }
        }
        httpChannel.cancel(Components.results.NS_BINDING_ABORTED);
    } 

    var statusCode = httpChannel.responseStatus;
    if (statusCode == "300" || statusCode == "301" || statusCode == "302" || statusCode == "303" || statusCode == "307"){
      //debugMsg("redirect spotted: " + statusCode + "\n");
      
      // Viva Duck-Typing :)
      location = httpChannel.getResponseHeader("location");
      var myContentLocation = new Array();
      myContentLocation["asciiSpec"] = location;
      
      if (this.rr.shouldLoad("",myContentLocation,httpChannel.originalURI,"redirect","","") != nsIContentPolicy.ACCEPT){        
			httpChannel.cancel(Components.results.NS_BINDING_ABORTED);
      }
    }
  }
}

//class constructor
function LocalRodeo() {
};

// class definition
LocalRodeo.prototype = {  
    mPrefBranch: Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefService).getBranch("extensions.requestrodeo.").QueryInterface(Components.interfaces.nsIPrefBranch2),   
  _enabled: function() {
    try {
        enabled = this.mPrefBranch.getBoolPref('localrodeo.enabled');       
    } catch (e) {
        enabled = false; 
    }
    return enabled;
  },
  /* parsing urls */
  //_getHostname: /\w+:\/\/([\w\.-]+)(:\d+)?/,
  _getHostname: /^\w+:\/\/([^@]*@)?([\w\.-]+)(:\d+)?/,
  getHostname: function(url)
  {
    var parsed = this._getHostname.exec(url);
    if (parsed)
      return parsed[2];
    
    /* malformed URL */
    return false;
  },
  
  /* we need to determine if a host is in our local subnet
     Private subnets are local in the default setting */
  _isLocalRegex: [ /^127\.0\.0\.1/, /^192\.168\./, /^169\.254\./,
		   /^172\.(?:(?:1[6789])|(?:2[0123456789])|(?:3[01]))\./,
		   /^10\./ ],
//		   /^134\.100\./ ],
  /* the resolver xpcom */
  _dnsService: Components.classes["@mozilla.org/network/dns-service;1"].
                                getService(Components.interfaces.nsIDNSService),
  resolve: function (host){
    /* if we dont get an answer, we just pass false on */
    try{
      //return this._dnsService.resolve(host,
      //	    Components.interfaces.nsIDNSService.RESOLVE_BYPASS_CACHE);
      var res = this._dnsService.resolve(host,0);
      return res; //enable caching
    }
    catch (e) {}
    
    return false;
  },
  
  /* Check if the passed argument is an ip-address*/ 
  isIP: function (host){
    var _isIP = /^(\d\d?\d?)\.(\d\d?\d?)\.(\d\d?\d?)\.(\d\d?\d?)$/;
    var res = _isIP.exec(host)
    if (res)
      return true;
    else 
      return false;
  },
  
  /* lookup without cache */
  isLocal: function (host) {
    //debugMsg('isLocal('+host+') called.\n');
    
    var result = false;
    
    /* if host is null assume that its non local */
    if (!host){
      result =  "unknown";
    } else if (this.isIP(host)){
      for (var i in this._isLocalRegex)
	result = result || host.match(this._isLocalRegex[i]) != null;
    } else {
      
      var dnsresult = this.resolve(host);
      if (dnsresult) {
		var ip;
		do {
	  
		  ip = dnsresult.getNextAddrAsString();
	  	  //debugMsg('Got IP: ' + ip + ' and is this local? ' +
		  //   (ip.match(this._isLocalRegex[i]) != null ? 'true' : 'false') + '\n');
	  
		  for (var i in this._isLocalRegex)
		    result = result || ip.match(this._isLocalRegex[i]) != null;
	
		} while (dnsresult.hasMore());
      
      } else {
		// DNS lookup failed
		result = "unknown"; 
      }
    }
    //debugMsg('isLocal returned ' + result + '\n');
    return result;
  },

  /* our cache */
  _isLocal: new Object(),

  /* this is the lookup function using our cache. note that its necessary
     to do a lookup even if we have a cached result in case this is a lookup
     for an request target */
  isLocalCached: function (host, lookupForTarget){
    /* if host is null (failed lookup) assume that its non local */
    if (!host)
      return false;

    var result = false;

    if (typeof(this._isLocal[host]) == 'boolean'){
        oldEntry = this._isLocal[host]; 
        /* it is in our cache */
        //debugMsg('Cache hit.\n');
		if (lookupForTarget && !this._isLocal[host]){
	 	 /* do a fresh lookup if cache says its remote */
	  	 result = this.isLocal(host);
	  
	  	// Check for anti-pinning attacks
	  	if (oldEntry != result){
	    	logToConsole("ANTI-DNS-PINNING ATTACK spotted, adding host "+ host +" to blacklist.");
	    	antiPinnerList[host] = true;  
	  	}
      } else
	  result = this._isLocal[host];
    } else {
		/* it is not... lets look it up */
		//debugMsg('Cache miss.\n');
		result = this.isLocal(host);
		this._isLocal[host] = result;
  	}
    
    //debugMsg('Host ' + host + ' is ' + (result ? 'local.' : 'extern.') + '\n');
    return result;
  },
  
  shouldLoad: function (contentType, contentLocation, requestOrigin, context,
                        mimeTypeGuess, extra){
    
    if (!this._enabled()){
      // all requests are permitted
      return nsIContentPolicy.ACCEPT;
    } 
    
    // For the lack of a better place, we instantiate our observer here, if it does not extist yet  
    if (!ro){
        ro = new RodeoObserver(this);
        ro.addToListener();
    }
    
    /* Check for chrome://-URLs */
    /*if (contentLocation.asciiSpec.match(/^chrome/)){
      debugMsg("\nChrome found: ");
      debugMsg("\n  From:    " + requestOrigin.asciiSpec); 
      debugMsg("\n  To:      " + contentLocation.asciiSpec); 
      debugMsg("\n  Context: " + context); 
      debugMsg("\n\n");
      }*/

    /* accept all non-network requests */
    if (!(contentLocation.asciiSpec.match(/^http/) ||
          contentLocation.asciiSpec.match(/^ftp/)))
      return nsIContentPolicy.ACCEPT;
    
    /* Requests from or to hosts, that are part of a ant-DNS-pinning attack are denied */
    if (antiPinnerList[this.getHostname(requestOrigin.asciiSpec)]){
        debugMsg("Request from a host that is part of a anti-DNS-Pinning attack denied\n");
        return nsIContentPolicy.REJECT_SERVER;
    }
    if (antiPinnerList[this.getHostname(contentLocation.asciiSpec)]){
        debugMsg("Request to a host that is part of a anti-DNS-Pinning attack denied\n");
        return nsIContentPolicy.REJECT_SERVER;
    }
    
    /* big fat log entry */
    //debugMsg('shouldLoad(');
    args = [contentType, contentLocation.asciiSpec, requestOrigin.asciiSpec,
            context, mimeTypeGuess, extra];
    //for (p in args){
    //  debugMsg(args[p] + ', ');
    //}
    //debugMsg(') called.\n'); 
       
    /* is this a request initated by user interaction with the browser gui? */
    if (requestOrigin.asciiSpec == 'chrome://browser/content/browser.xul')
      return nsIContentPolicy.ACCEPT;
        
    var fromHost = this.getHostname(requestOrigin.asciiSpec);
    var toHost = this.getHostname(contentLocation.asciiSpec);
    var isFromHostLocal = this.isLocalCached(fromHost);
    var isToHostLocal = this.isLocalCached(toHost, true);
    
    /* be verbose */
    //debugMsg(requestOrigin.asciiSpec + " : "  + fromHost + ' isLocal? ' + isFromHostLocal + '\n');
    //debugMsg(contentLocation.asciiSpec + ": " + toHost   + ' isLocal? ' + isToHostLocal   + '\n');
    
    var forbidden = true;
    
    // We allow all requests with a local origin
    if (isFromHostLocal == true) {
        forbidden = false; 
	
    // We allow requests from unknown origin only if they have a remote target
    } else if (isFromHostLocal == "unknown") {
        if (isToHostLocal == false)
            forbidden = false; 

    // We allow requests to a unknow target only under certain conditions (depending on the 'context')
    // The reason is, that we cannot differntiate between obfusication attacks and failed DNS-lookups
    // Therefore, when it is save, we allow the request 
    // Otherwise we do notification via the error console
    } else if (isToHostLocal == "unknown") {

      // [object XULElement] means that a link was clicked and the result will load in the top of a window
      if (context == "[object XULElement]"){
			forbidden = false; 
      
      // In the case of an iFrame we remove onload and onerror events
      } else if (context == "REMOVEME[object XPCNativeWrapper [object HTMLIFrameElement]]"){
		context.setAttribute("onload", "");
		context.setAttribute("onerror", "");      
		forbidden = false; 

      // all other cases (image tag, script tag, link tag, etc.) are not allowed
      // In this case we warn through the error console
      } else {
		logToConsole("Unknown destination: " + this.getHostname(contentLocation.asciiSpec) + ". Request stopped."); 
      }

    // We allow requests that have both a remote origin and target
    } else if (isFromHostLocal == false){
        if (isToHostLocal == false)
            forbidden = false; 
    }

    // Every thing else is FORBIDDEN
        
    if (forbidden){
      /* neat ui integration has to wait */
      if (isToHostLocal == true)
		message = 'Blocking request to local resource (' + contentLocation.asciiSpec + ').';
      else
		message = 'Blocking request to unknown resource (' + contentLocation.asciiSpec + ').';
      	message = message + '\nRequest-origin was: '+ requestOrigin.asciiSpec; 
      logToConsole(message);

      return nsIContentPolicy.REJECT_SERVER;
    }

    return nsIContentPolicy.ACCEPT;
  },

  shouldProcess: function (contentType, contentLocation, requestOrigin, context,
                           mimeType, extra){
    return nsIContentPolicy.ACCEPT;
  },

  observe: function (aSubject,  
		     aTopic,  
		     aData)
  {
    /* we registered at xpcom-startup. this might be useful in the future */
  },

  QueryInterface: function(aIID)
  {
    if (!aIID.equals(nsIContentPolicy) &&
	!aIID.equals(nsIObserver) &&
	!aIID.equals(nsISupports))
      throw Components.results.NS_ERROR_NO_INTERFACE;
    return this;
  }  
  
};

/***********************************************************
class factory

This object is a member of the global-scope Components.classes.
It is keyed off of the contract ID. Eg:

myLocalRodeo = Components.classes['@requestrodeo/contentpolicy;1'].
                         createInstance(Components.interfaces.nsIContentPolicy);

***********************************************************/
var LocalRodeoFactory = {
  createInstance: function (aOuter, aIID)
  {
    if (aOuter != null)
      throw Components.results.NS_ERROR_NO_AGGREGATION;
       
    logToConsole("Booting up LocalRodeo");
    rr = new LocalRodeo();
    return rr.QueryInterface(aIID);
  }
};

/***********************************************************
module definition (xpcom registration)
***********************************************************/
var LocalRodeoModule = {
  _firstTime: true,
  registerSelf: function(aCompMgr, aFileSpec, aLocation, aType)
  {
    /* debugMsg('registerSelf()\n'); */
    aCompMgr = aCompMgr.
        QueryInterface(Components.interfaces.nsIComponentRegistrar);
    aCompMgr.registerFactoryLocation(CLASS_ID, CLASS_NAME, 
        CONTRACT_ID, aFileSpec, aLocation, aType);

    /* and register it */
    var catman = Components.classes['@mozilla.org/categorymanager;1'].
                           getService(Components.interfaces.nsICategoryManager);

    catman.addCategoryEntry('xpcom-startup',
			    CONTRACT_ID,
			    CONTRACT_ID,
			    true,
			    true);

    catman.addCategoryEntry('content-policy',
			    CONTRACT_ID,
			    CONTRACT_ID,
			    true,
			    true);
  },

  unregisterSelf: function(aCompMgr, aLocation, aType)
  {
    aCompMgr = aCompMgr.
        QueryInterface(Components.interfaces.nsIComponentRegistrar);
    aCompMgr.unregisterFactoryLocation(CLASS_ID, aLocation);        

    var catman = Components.classes['@mozilla.org/categorymanager;1'].
                           getService(Components.interfaces.nsICategoryManager);

    catman.deleteCategoryEntry('xpcom-startup',
			       CONTRACT_ID,
			       PR_TRUE);

    catman.deleteCategoryEntry('content-policy',
			       CONTRACT_ID,
			       PR_TRUE);
  },
  
  getClassObject: function(aCompMgr, aCID, aIID)
  {
    if (!aIID.equals(Components.interfaces.nsIFactory))
      throw Components.results.NS_ERROR_NOT_IMPLEMENTED;

    if (aCID.equals(CLASS_ID))
      return LocalRodeoFactory;

    throw Components.results.NS_ERROR_NO_INTERFACE;
  },

  canUnload: function(aCompMgr) { return true; }
};

/***********************************************************
module initialization

When the application registers the component, this function
is called.
***********************************************************/
function NSGetModule(aCompMgr, aFileSpec)
{
  /* debugMsg('NSGetModule called.\n'); */

	// Initialise Observers here (?)
	// ...
	
    return LocalRodeoModule;
}


function objDumper(name, obj){
  /*debugMsg("\nContents of " + name + ": "+ obj + " (" + typeof(obj) + ")\n");
    for (i in obj){
        debugMsg("->  "+i+":\t"+obj[i]+"\n");
    }
    debugMsg("\n");*/
}

var consoleService = Components.classes["@mozilla.org/consoleservice;1"].getService(Components.interfaces.nsIConsoleService);

function logToConsole (message){
  consoleService.logStringMessage("LocalRodeo: " + message);
  //debugMsg(message + "\n");
}

var mPrefBranch = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefService).getBranch("extensions.requestrodeo.").QueryInterface(Components.interfaces.nsIPrefBranch2);

function debugMsg(message){
  	debug = mPrefBranch.getBoolPref('localrodeo.debug');       

	if (debug){
    	logToConsole(message);
    	dump("Localrodeo: " + message + "\n");  // Dumps to commandline io 
	}
}
