/*
** LocalRodeo - Client Side Protection against Session Riding
** Copyright (C) 2006 Justus Winter <4winter@informatik.uni-hamburg.de>
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 2
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
** MA  02110-1301, USA
*/

var oLocalRodeo;

function enableRR(){
  if (!oLocalRodeo){
      oLocalRodeo = new LocalRodeo();
  }
  //oLocalRodeo.addToListener();
}

function disableRR(){
  if (oLocalRodeo){
      //oLocalRodeo.removeFromListener();
  }
}

function LocalRodeo()
{
  /* watch our options panel */
  this.mPrefBranch.addObserver("", this, false);

  this._statusBarPanel = document.getElementById('requestrodeo-panel');

  /* create instance of our xpcom contentpolicy manager */
  //requestRodeoContentPolicy = Components.classes['@requestrodeo/contentpolicy;1'].createInstance(Components.interfaces.nsIContentPolicy);
}

LocalRodeo.prototype =
{
  /* options */
  mPrefBranch: Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefService).getBranch("extensions.requestrodeo.").QueryInterface(Components.interfaces.nsIPrefBranch2),
  
  showOptionsDialog: function (){
    window.openDialog("chrome://localrodeo/content/options.xul", 
		      "options-dialog",
		      "chrome,dialog,centerscreen,resizable=no");
  },

  /* parsing urls */
  _parseURL: /(\w+):\/\/([\w\.-]+)(:\d+)?(.*)/,
  parseURL: function(url)
  {
    var parsed = this._parseURL.exec(url);
    if (parsed)
      {
	var result = new Object();
	result.toString = function()
	  {
	    return url;
	  };
	result.protocol = parsed[1];
	result.hostname = parsed[2];
	result.port = parsed[3] ||
	  result.protocol == 'http' ? 80 : false ||
	  result.protocol == 'https' ? 443 : false;
	result.resource = parsed[4];
	return result;
    }
    
    /* malformed URL */
    return false;
  },
  
  /* debugfoo */
  debug: true,
  _consoleService: Components.classes["@mozilla.org/consoleservice;1"].getService(Components.interfaces.nsIConsoleService),
  log: function(message)
  {
    if (!this.debug)
      return;

    if (this.mPrefBranch.getBoolPref('logToConsoleService'))
      this._consoleService.logStringMessage("LocalRodeo: " + message);

    if (this.mPrefBranch.getBoolPref('logToConsole'))
      dump("LocalRodeo: " + message + '\n');
  },

  /* This is the cut down observerService's observe listener from TamperData */
  observe: function(aSubject, aTopic, aData)
  {
    if (aTopic == 'http-on-modify-request')
      {
      	
	aSubject.QueryInterface(Components.interfaces.nsIHttpChannel);
	this.onModifyRequest(aSubject);
      }
    else
    if (aTopic == 'http-on-examine-response')
      {
        /* if we prevented some csrf, display a message */
	aSubject.QueryInterface(Components.interfaces.nsIHttpChannel);
	try
	{
	  this.displayNotification(aSubject.getRequestHeader('X-LocalRodeo'));
	}
	catch (e)
	{ }
      }
  },

  /* examine request */
  onModifyRequest : function (oHttp)
  {
    var isEntitled = true;
    var reason = '';
    var uri = this.parseURL(oHttp.URI.asciiSpec);
    var referrer = this.parseURL(oHttp.referrer && oHttp.referrer.asciiSpec);

    /* log request */
    this.log('got an onModifyRequest. uri is "' + uri + '", ' +
             'referrer is "' + referrer + '",' +
             'originalURI is ' + oHttp.originalURI.asciiSpec + '.');

    /* for some strange reason, requests to favicons do not carry an referrer */
    if (oHttp.URI.asciiSpec == (uri.protocol + '://' + uri.hostname + '/favicon.ico'))
      {
        this.log('Requests to favicons are harmless.');
        return;
      }

    /* no referrer? then its not entitled */
    if (!referrer){
		reason += 'No referrer.'
		isEntitled = false;
    }

    /* netlocs different? */
    if (referrer &&
	((uri.hostname != referrer.hostname) ||
	 (uri.port != referrer.port))){
		reason += 'The netlocs differ.';
		isEntitled = false;
    }

    oHttp.setRequestHeader('X-Protection', 'LocalRodeo extension', false);

    if (!isEntitled){
		this.removeCredentials(oHttp, reason);
    }
  },

  /* inform the user */
  displayNotification: function (message, delayed){
    /* wait for the browser to setup the 'window' */
    if (!delayed){
      window.setTimeout(this.displayNotification, 500, message, true);
      return;
   }

   if (gBrowser.getNotificationBox)
    {
      /* FF2 or newer */
      var notification = gBrowser.getNotificationBox();
      if (notification)
      {
        /* remove any previous notification we made */
        var n = notification.getNotificationWithValue('requestrodeo');
        if (n)
          notification.removeNotification(n);
      
        /* and display a new one */
        notification.appendNotification(message, 'requestrodeo',
                                        'chrome://browser/skin/Info.png',
                                        notification.PRIORITY_WARNING_MEDIUM,
                                        null);
      }
      else
        this.log('Still no notification area.');
    }
    else
    if (gBrowser.showMessage)
    {
      /* FF1.x */
      gBrowser.showMessage(gBrowser.selectedBrowser, 'chrome://browser/skin/Info.png',
                           message, 'DoNothing', null, null, '',
                           'top', true, null);
    }
    else
    {
      /* just log it */
      oLocalRodeo.log(message);
    }
  },
  displayNotificationOnResponse: function (message, request)
  {
    request.setRequestHeader('X-LocalRodeo', message, false);
  },
  
  /* remove any implicit authentication credentials */
  removeCredentials: function (request, reason)
  {
    this.log('removeCredentials(' + reason + ')');
    // TODO: remove http-auth, client side ssl auth

    var removedHeaders = new Array();
    var headers = ['Cookie', 'Cookie2'];
    for (k in headers)
    {
      try
      {
        var h = headers[k];
        /* this will fail if there is no header h */
        request.getRequestHeader(h);
        request.setRequestHeader(h, '', false);
        removedHeaders.push(h);
      } catch (e) {}
    }
    
    /* inform the user */
    if (removedHeaders.length != 0)
      this.displayNotificationOnResponse('LocalRodeo prevented a possible Session Riding attempt by removing the Header ' + removedHeaders + '. Reason: ' + reason,
                                         request);
  },

  /* from TamperData */
  _observerService: Components.classes["@mozilla.org/observer-service;1"].getService(Components.interfaces.nsIObserverService),
  addToListener: function()
  {
    this._observerService.addObserver(this, "http-on-modify-request",   false);
    this._observerService.addObserver(this, "http-on-examine-response", false);
    this.log('LocalRodeo enabled.');
    this._statusBarPanel.tooltipText = 'LocalRodeo enabled.';
    this._statusBarPanel.style.color = 'black';
  },

  /* from TamperData */
  removeFromListener: function()
  {
    this._observerService.removeObserver(this, "http-on-modify-request");
    this._observerService.removeObserver(this, "http-on-examine-response");
    this.log('LocalRodeo disabled.'); 
    this._statusBarPanel.tooltipText = 'LocalRodeo disabled.';
    this._statusBarPanel.style.color = 'grey';
  }
};
